# ------------------------------------------------------------------------------

# Testes de independência com as medidas de Pearson, Speraman, Kendall, 
# Hoeffding, HHG, CIM, IM e Dcor

# ------------------------------------------------------------------------------


# ================================ BIBLIOTECAS ================================= 

require(Hmisc) # pacote com a implementação da medida D de Hoeffding
require(energy) # pacote com a implementação da correlação de distância
require(HHG2x2) # pacote com a implementação da medida HHG
require(entropy) # para implementação da informação mútua
library(gregmisc)
source("MINE.r") # script para executar o programa que calcula o MIC


# ============================ FUNCOES DA INTERFACE ============================ 

# Pearson ----------------------------------------------------------------------

# Devolve o coeficiente de Pearson entre as amostras x e y.
pearson <- function (x, y) {
    n <- length(x)
    sumx <- 0
    sumy <- 0
    sumx2 <- 0
    sumy2 <- 0
    xy <- 0
    
    for (i in 1:n) {
        xy <- xy + x[i]*y[i]
        sumx <- sumx + x[i]
        sumy <- sumy + y[i]
        sumx2 <- sumx2 + x[i]*x[i]
        sumy2 <- sumy2 + y[i]*y[i]
    }

    r <- n*xy - sumx*sumy
    r <- r/(sqrt(n*sumx2 - sumx*sumx)*sqrt(n*sumy2 - sumy*sumy))
    return(r)
}

# Realiza um teste de independencia entre as variaveis de amostras x e y, 
# utilizando o coeficiente de Pearson. 
#
# Devolve uma lista com o valor da estatistica de teste, os graus de liberdade
# da estatistica, o p-valor do teste e o coeficiente de Pearson estimado entre 
# as variaveis
pearson.test <- function (x, y) {
    return(correlation.test(x, y, "pearson"))
}

# Spearman ---------------------------------------------------------------------

# Devolve o coeficiente de Spearman entre as amostras x e y.
spearman <- function (x, y) {
    ix_x <- sort.list(x)
    ix_y <- sort.list(y)
    
    n <- length(x)
    
    d <- rep(0, times=n)
    
    # Obs: Estamos supondo que nao existe i!=j tal que (xi = xj) ou (yi = yj)
    
    # Diferenças entre os postos dos valores de x e y:
    d <- rank(x)
    for (i in 1:n)
        d[ix_y[i]] <- d[ix_y[i]] - i

    sumd2 <- 0

    for (i in 1:n)
        sumd2 <- sumd2 + d[i]*d[i]

    return(1 - (6*sumd2)/(n*(n*n - 1)))
}

# Realiza um teste de independencia entre as variaveis de amostras x e y 
# utilizando a correlação de Kendall. 
#
# Devolve uma lista com o valor da estatistica de teste, os graus de liberdade
# da estatistica, o p-valor do teste e o coeficiente de Spearman estimado entre 
# as variaveis
spearman.test <- function (x, y) {
    # Aproximacao para amostras de comprimento maior que 20
    #return(correlation.test(x, y, "spearman"))
    result <- cor.test(x, y, method = "spearman") 
    
    return(list("coeficiente" = as.numeric(result["estimate"]), 
                "estatistica" = as.numeric(result["statistic"]),
                "p-valor" = as.numeric(result["p.value"])))
}

# Kendall ----------------------------------------------------------------------

# Devolve a correlacao de distancia entre as amostras x e y
kendall <- function (x, y) {
   return(cor(x, y, method = "kendall"))
}

# Realiza um teste de independencia entre as variaveis de amostras x e y 
# utilizando a correlacao de distancia. Temos:
# H0: As variaveis sao independentes
# H1: As variaveis nao sao independentes
#
# Devolve uma lista com o tau de kendall, o valor da estatistica do 
# teste e o p-valor
kendall.test <- function (x, y) {
    result <- cor.test(x, y, method = "kendall") 
    
    return(list("tau" = as.numeric(result["estimate"]), 
                "estatistica" = as.numeric(result["statistic"]),
                "p-valor" = as.numeric(result["p.value"])))
}

# Informação mútua -------------------------------------------------------------

# Devolve a informação mútua entre as amostras x e y
mutualinfo <- function(x, y) {
    Mx <- matrix(x, length(x), 1)
    My <- matrix(y, length(y), 1)
    nbins <- ceiling(log(length(Mx[,1]),2)) + 1
    a <- hist2d(cbind(Mx,My), nbins=nbins,show=FALSE)
    result <- mi.empirical(a$counts)
    return(result)
}

# Realiza um teste de independencia entre as variaveis de amostras x e y 
# utilizando informação mútua. Temos:
# H0: As variaveis sao independentes
# H1: As variaveis nao sao independentes
# 
# Devolve uma lista com o a informação mútua e o p-valor que é 
# estimado utilizando-se nresamplings reamostragens
mutualinfo.test <- function(x, y, nresamplings = 1000) {
    rho <- mutualinfo(x, y) 
    rhos <- vector(length = nresamplings)
    for (i in 1:nresamplings)
        rhos[i] <- mutualinfo(x, sample(y, replace = FALSE))
    pvalue <- 1 - rank(c(rho, rhos, recursive=TRUE), ties="first")[1]/(nresamplings + 1)
    return(list("mutualinfo" = rho, 
                "p-valor" = pvalue))
}

# Hoeffding --------------------------------------------------------------------

# Devolve a medida D de Hoeffding entre as amostras x e y
hoeffding <- function (x, y) {
    n = length(x)
    Q <- rep(0, times = n)

    R <- rank(x)
    S <- rank(y)

    for (i in 1:n) {
        Q[i] <- 0
        for (j in 1:n)
            if (x[j] < x[i] && y[j] < y[i])
                Q[i] <- Q[i] + 1
    }

    D1 <- 0
    D2 <- 0
    D3 <- 0

    for (i in 1:n) {
        D1 <- D1 + Q[i]*(Q[i] - 1)
        D2 <- D2 + (R[i] - 1)*(R[i] - 2)*(S[i] - 1)*(S[i] - 2)
        D3 <- D3 + (R[i] - 2)*(S[i] - 2)*Q[i]
    }
    
    D <- (n - 2)*(n - 3)*D1 + D2 - 2*(n - 2)*D3
    D <- D/(n*(n - 1)*(n - 2)*(n - 3)*(n - 4))

    return(D)
}

# Realiza um teste de independencia entre as variaveis de amostras x e y 
# utilizando a medida D de Hoeffding. Temos:
# H0: As variaveis sao independentes
# H1: As variaveis nao sao independentes
#
# Devolve uma lista com a medida D de Hoeffding, o nivel de significancia do
# teste e o valor de rho (rejeitamos H0 se 30*D > rho (?))
hoeffding.test <- function (x, y, alpha = 0.05) {
    D <- hoeffding(x, y)
    n <- length(x)
    
    # TODO : Achar uma forma mais bonita de extrair o p-valor da saida de hoeffd
    pvalue <- c(hoeffd(x, y)["P"], recursive = TRUE)[2]
    
    rho <- sqrt((2*(n*n + 5*n - 32))/(9*n*(n - 1)*(n - 3)*(n - 4)*alpha))
    
    return(list("D" = D, "nivel de significancia" = alpha, "rho" = rho,
                "p-valor" = pvalue))
}

# Correlacao de distancia ------------------------------------------------------

# Devolve a correlacao de distancia entre as amostras x e y
dcorrelation <- function (x, y) {
   return(dcor(x, y))
}

# Realiza um teste de independencia entre as variaveis de amostras x e y 
# utilizando a correlacao de distancia. Temos:
# H0: As variaveis sao independentes
# H1: As variaveis nao sao independentes
#
# Devolve uma lista com a correlacao de distancia,o valor da estatistica do 
# teste e o p-valor
dcorrelation.test <- function (x, y, nresamplings = 1000) {
    result <- dcov.test(x, y, R = nresamplings) 
    
    return(list("correlacao de distancia" = as.numeric(result["estimate"]), 
                "estatistica" = as.numeric(result["statistic"]),
                "p-valor" = as.numeric(result["p.value"])))
}

# Coeficiente de Informação Máxima ---------------------------------------------

# Devolve o coeficiente de informação máxima entre as amostras x e y
mic <- function (x, y) {
   data <- matrix(0, 2, length(x))
   data[1, ] <- x
   data[2, ] <- y 
   rMINE(data, "mic", 0)
   result <- scan("mic,mv=0,cv=0.0,B=n^0.6,Results.csv",what="",sep=",")
   return(as.numeric(result[11]))
}

# Realiza um teste de independencia entre as variaveis de amostras x e y 
# utilizando o coeficiente de informação máxima. Temos:
# H0: As variaveis sao independentes
# H1: As variaveis nao sao independentes
# 
# Devolve uma lista com o coeficiente de informação máxima e o p-valor que é 
# estimado utilizando-se nresamplings reamostragens
mic.test <- function (x, y, nresamplings = 1000) {
    rho <- mic(x, y) 
    rhos <- vector(length = nresamplings)
    for (i in 1:nresamplings)
        rhos[i] <- mic(x, sample(y, replace = FALSE))
    pvalue <- 1 - rank(c(rho, rhos, recursive=TRUE), ties="first")[1]/(nresamplings + 1)
    return(list("cim" = rho, 
                "p-valor" = pvalue))
}

# HHG --------------------------------------------------------------------------

# Devolve a medida HHG entre as amostras x e y
hhg <- function (x, y) {

}

# Realiza um teste de independencia entre as variaveis de amostras x e y 
# utilizando a medida HHG. Temos:
# H0: As variaveis sao independentes
# H1: As variaveis nao sao independentes
#
# Devolve uma lista com o  p-valor
hhg.test <- function (x, y, nresamplings = 1000) {
    Dx <- as.matrix(dist(x, diag=TRUE, upper=TRUE))
    Dy <- as.matrix(dist(y, diag=TRUE, upper=TRUE))
    pvalue <- pvHHG(Dx, Dy, alpha_hyp=1, monte = nresamplings)$pv
    return(list("p-valor" = pvalue))
}


# ============================= FUNCOES AUXILIARES ============================= 

# Realiza um teste de independencia entre as variaveis de amostras x e y 
# utilizando o coeficiente de Pearson ou Spearman. Temos:
# H0: As variaveis sao independentes
# H1: As variaveis nao sao independentes
#
# Devolve uma lista com o valor da estatistica de teste, os graus de liberdade
# da estatistica, o p-valor do teste e o coeficiente de Pearson ou Spearman 
# estimado entre as variaveis
correlation.test <- function (x, y, method="pearson") {
    if (method == "pearson")
        r <- pearson(x, y)
    else if (method == "spearman")
        r <- spearman(x, y) 
    n <- length(x)
    # Sob H0, t segue uma distribuicao t de Student com n - 2 graus de 
    # liberdade:
    if (round(1 - r*r, digits = 10) == 0)
        t <- Inf
    else
       t <- r*sqrt(n - 2)/sqrt(1 - r*r) 
    # O p-valor e dado pela probabilidade de se obter uma estatistica de teste
    # igual ou mais extrema que aquela observada na amostra (t), assumindo que
    # H0 e verdadeira:
    if (r >= 0)
        # Probabilidade de a estatistica do teste ser maior ou igual a t
        pval <- 2*pt(-t, n - 2)
    else
        # Probabilidade de a estatistica do teste ser menor ou igual a t:
        pval <- 2*pt(t, n - 2)
    return(list("estatistica" = t, "graus de liberdade" = n-2, "p-valor" = pval,
                "coeficiente" = r))
}
